//
// Copyright (c) Microsoft. All rights reserved.
// See https://aka.ms/azai/vision/license for the full license information.
//

#pragma once
#include <azac_api_cxx_details_json.h>
#include <vision_api_cxx_common.h>
#include <vision_api_cxx_details_result_base.h>
#include <vision_api_cxx_image_analysis_result_properties.h>
#include <vision_api_cxx_image_analysis_result_reason.h>
#include <vision_api_cxx_image_analysis_core_result_reason.h>
#include <vision_api_cxx_image_analysis_core_stop_reason.h>

namespace Azure {
namespace AI {
namespace Vision {
namespace ImageAnalysis {

using namespace Azure::AI::Core::Details::Json;
using namespace Azure::AI::Vision::Core::Details;
using namespace Azure::AI::Vision::ImageAnalysis::Details;

/// <summary>
/// Represents a rectangular area inside an image.
/// </summary>
class Rectangle
{
public:
    /// <summary> 
    /// X-coordinate of the top left point of the area, in pixels.
    /// </summary>
    int32_t X = 0;

    /// <summary> 
    /// Y-coordinate of the top left point of the area, in pixels.
    /// </summary>
    int32_t Y = 0;

    /// <summary> 
    /// Height measured from the top-left point of the area, in pixels.
    /// </summary>
    int32_t H = 0;

    /// <summary>
    /// Width measured from the top-left point of the area, in pixels.
    /// </summary>
    int32_t W = 0;

    /// <summary>
    /// Helper method to display the Rectangle values
    /// </summary>
    /// <returns>
    /// A string containing the values of X, Y, H and W.
    /// </returns>
    std::string ToString() const
    {
        return "{X=" + std::to_string(X) +
               ",Y=" + std::to_string(Y) +
               ",Width=" + std::to_string(W) +
               ",Height=" + std::to_string(H) + "}";
    }

public:
    /*! \cond PRIVATE */

    /// <summary>
    /// Construct the Rectangle object from the appropriate JSON element in the service response.
    /// </summary>
    Rectangle(JsonValue jsonBoundingBox)
    {
        X = (int32_t)(jsonBoundingBox["x"].AsInt());
        Y = (int32_t)(jsonBoundingBox["y"].AsInt());
        W = (int32_t)(jsonBoundingBox["w"].AsInt());
        H = (int32_t)(jsonBoundingBox["h"].AsInt());
    }

    /// <summary>
    /// Construct the Rectangle object from given x, y, width and height values
    /// </summary>
    Rectangle(int32_t x, int32_t y, int32_t w, int32_t h)
    {
        X = x;
        Y = y;
        W = w;
        H = h;
    }

    Rectangle() = default;

    /*! \endcond */
};

/// <summary>
/// Represents a physical object detected in an image.
/// </summary>
class DetectedObject
{
public:
    /// <summary>
    /// A label that briefly describes the detected object.
    /// </summary>
    std::string Name;

    /// <summary>
    /// A score that represents the likelihood that this detection was accurate.
    /// </summary>
    /// <remarks>
    /// Confidence scores span the range of 0.0 to 1.0 (inclusive), with higher values indicating higher
    /// probability.
    /// </remarks>
    double Confidence = 0;

    /// <summary>
    /// A rectangular boundary within which the object was detected.
    /// </summary>
    /// <remarks>
    /// Coordinates are are in pixels, with (0,0) being the top-left of the source image.
    /// </remarks>
    Rectangle BoundingBox;

public:
    /*! \cond PRIVATE */

    /// <summary>
    /// Construct the DetectedObject from the appropriate JSON element in the service response.
    /// </summary>
    DetectedObject(JsonValue jsonObjectResult)
    {
        Name =  jsonObjectResult["tags"][0]["name"].AsString();
        Confidence =  jsonObjectResult["tags"][0]["confidence"].AsDouble();
        BoundingBox = Rectangle(jsonObjectResult["boundingBox"]);
    }

    /*! \endcond */
};

/// <summary>
/// Represents a list of physical object detected in an image.
/// </summary>
class DetectedObjects: public std::vector<DetectedObject>
{
public:
    /*! \cond PRIVATE */

    /// <summary>
    /// Construct the DetectedObjects from the appropriate JSON element in the service response.
    /// </summary>
    DetectedObjects(JsonValue jsonObjectsResult)
    {
        if (jsonObjectsResult.HasValue("values"))
        {
            JsonValue jsonObjectsResultValues = jsonObjectsResult["values"];

            for (int i = 0; i < jsonObjectsResultValues.Count(); ++i)
            {
                push_back(DetectedObject(jsonObjectsResultValues[i]));
            }
        }
    }

    /*! \endcond */
};

/// <summary>
/// Represent an image tag. A tag can be a recognizable object, living being, scenery, or actions that appear in the image.
/// </summary>
class ContentTag
{
public:
    /// <summary>
    /// Name of the tag.
    /// </summary>
    /// <remarks>
    /// Note that you can control the tag language by calling ImageAnalysisOptions::SetLanguage().
    /// </remarks>
    std::string Name;

    /// <summary>
    /// A score that represents the likelihood that this detection was accurate.
    /// </summary>
    /// <remarks>
    /// Confidence scores span the range of 0.0 to 1.0 (inclusive), with higher values indicating higher
    /// probability.
    /// </remarks>
    double Confidence = 0;

public:
    /*! \cond PRIVATE */

    /// <summary>
    /// Construct the ContentTag object from the appropriate JSON element in the service response.
    /// </summary>
    ContentTag(JsonValue jsonTagsResultValue)
    {
        Name = jsonTagsResultValue["name"].AsString();
        Confidence = jsonTagsResultValue["confidence"].AsDouble();
    }

    /*! \endcond */
};

/// <summary>
/// Represents a list of image tags
/// </summary>
class ContentTags: public std::vector<ContentTag>
{
public:
    /*! \cond PRIVATE */

    /// <summary>
    /// Construct the ContentTags object from the appropriate JSON element in the service response.
    /// </summary>
    ContentTags(JsonValue jsonTagsResult)
    {
        if (jsonTagsResult.HasValue("values"))
        {
            JsonValue jsonTagsResultValues = jsonTagsResult["values"];

            for (int i = 0; i < jsonTagsResultValues.Count(); ++i)
            {
                push_back(ContentTag(jsonTagsResultValues[i]));
            }
        }
    }

    /*! \endcond */
};

/// <summary>
/// Represents a person detected in an image.
/// </summary>
class DetectedPerson
{
public:
    /// <summary>
    /// A score that represents the likelihood that this detection was accurate.
    /// </summary>
    /// <remarks>
    /// Confidence scores span the range of 0.0 to 1.0 (inclusive), with higher values indicating higher
    /// probability.
    /// </remarks>
    double Confidence = 0;

    /// <summary>
    /// A rectangular boundary within which the person was detected.
    /// </summary>
    /// <remarks>
    /// Coordinates are are in pixels, with (0,0) being the top-left of the source image.
    /// </remarks>
    Rectangle BoundingBox;

public:
    /*! \cond PRIVATE */

    /// <summary>
    /// Construct the DetectedPerson object from the appropriate JSON element in the service response.
    /// </summary>
    DetectedPerson(JsonValue jsonPeopleResultsValue)
    {
        Confidence = jsonPeopleResultsValue["confidence"].AsDouble();
        BoundingBox = Rectangle(jsonPeopleResultsValue["boundingBox"]);
    }

    /*! \endcond */
};

/// <summary>
/// Represents people detected in an image.
/// </summary>
class DetectedPeople: public std::vector<DetectedPerson>
{
public:
    /*! \cond PRIVATE */

    /// <summary>
    /// Construct the DetectedPeople object from the appropriate JSON element in the service response.
    /// </summary>
    DetectedPeople(JsonValue jsonPeopleResult)
    {
        if (jsonPeopleResult.HasValue("values"))
        {
            JsonValue jsonPeopleResultsValues = jsonPeopleResult["values"];

            for (int i = 0; i < jsonPeopleResultsValues.Count(); ++i)
            {
                push_back(DetectedPerson(jsonPeopleResultsValues[i]));
            }
        }
    }

    /*! \endcond */
};

/// <summary>
/// Represents a suggested image cropping that preserves much of the image content.
/// </summary>
class CropSuggestion
{
public:
    /// <summary>
    /// The aspect ratio of this crop suggestion
    /// </summary>
    /// <remarks>
    /// Aspect ratios are calculated by dividing the width of the cropped region by its height.
    ///
    /// You can request particular aspect ratios by calling the method ImageAnalysisOptions::SetCroppingAspectRatios.
    ///
    /// AspectRatio will be in the range 0.75 to 1.8 (inclusive) if ImageAnalysisOptions::SetCroppingAspectRatios was called,
    /// otherwise it will be in the range 0.5 to 2.0 (inclusive).
    /// </remarks>
    double AspectRatio = 0;

    /// <summary>
    /// Gets the rectangular bounds of the crop suggestion.
    /// </summary>
    /// <remarks>
    /// Coordinates are are in pixels, with (0,0) being the top-left of the source image.
    /// </remarks>
    Rectangle BoundingBox;

public:
    /*! \cond PRIVATE */

    /// <summary>
    /// Construct the CropSuggestion object from the appropriate JSON element in the service response.
    /// </summary>
    CropSuggestion(JsonValue jsonSmartCropsResultValue)
    {
        AspectRatio = jsonSmartCropsResultValue["aspectRatio"].AsDouble();
        BoundingBox = Rectangle(jsonSmartCropsResultValue["boundingBox"]);
    }

    /*! \endcond */
};

/// <summary>
/// Represents a list of image crop suggestions that preserve most of the image content
/// </summary>
class CropSuggestions: public std::vector<CropSuggestion>
{
public:
    /*! \cond PRIVATE */

    /// <summary>
    /// Construct the CropSuggestions object from the appropriate JSON element in the service response.
    /// </summary>
    CropSuggestions(JsonValue jsonSmartCropsResult)
    {
        if (jsonSmartCropsResult.HasValue("values"))
        {
            JsonValue jsonSmartCropsResultValues = jsonSmartCropsResult["values"];

            for (int i = 0; i < jsonSmartCropsResultValues.Count(); ++i)
            {
                push_back(CropSuggestion(jsonSmartCropsResultValues[i]));
            }
        }
    }

    /*! \endcond */
};

/// <summary>
/// Represents a generated phrase that describes the content of the whole image
/// or a region in the image.
/// </summary>
class ContentCaption
{
public:
    /// <summary>
    /// A generated phrase that describes the content of the whole image or a region in the image.
    /// </summary>
    std::string Content;

    /// <summary>
    /// A score, in the range of 0 to 1, representing the confidence that this description is accurate.
    /// </summary>
    /// <remarks>
    /// Confidence scores span the range of 0.0 to 1.0 (inclusive), with higher values indicating higher
    /// probability.
    /// </remarks>
    double Confidence = 0;

    /// <summary>
    /// A rectangular boundary to which the caption applies.
    /// </summary>
    /// <remarks>
    /// Coordinates are are in pixels, with (0,0) being the top-left of the source image.
    /// For the ImageAnalysisFeature::Caption result, this will be the whole image.
    /// For the ImageAnalysisFeature::DenseCaption result, this will either be the whole image
    /// or a region within the image.
    /// </remarks>
    Rectangle BoundingBox;

public:
    /*! \cond PRIVATE */

    /// <summary>
    /// Construct the ContentCaption object from the appropriate JSON element in the service response.
    /// </summary>
    ContentCaption(JsonValue jsonCaptionResult, int32_t imageWidth = 0, int32_t imageHeight = 0)
    {
        Content = jsonCaptionResult["text"].AsString();
        Confidence = jsonCaptionResult["confidence"].AsDouble();

        if (jsonCaptionResult.HasValue("boundingBox"))
        {
            // 'DenseCaptions' has bounding box values as part of JSON result. Parse it.
            BoundingBox = Rectangle(jsonCaptionResult["boundingBox"]);
        }
        else
        {
            // 'Caption' does not have bounding box values as part of JSON result. The bounding
            // box will be built with the provided input image width & height.
            BoundingBox = Rectangle(0, 0, imageWidth, imageHeight);
        }
    }

    /*! \endcond */
};

/// <summary>
/// Represents a list of up to 10 captions for different regions of the image.
/// </summary>
/// <remarks>
/// The first caption in the list represents the the whole image, and it is identical to the result
/// returned if you select the option ImageAnalysisFeature::Caption.
/// </remarks>
class DenseCaptions: public std::vector<ContentCaption>
{
public:
    /*! \cond PRIVATE */

    /// <summary>
    /// Construct the DenseCaptions from the appropriate JSON element in the service response.
    /// </summary>
    DenseCaptions(JsonValue jsonDenseCaptionsResult)
    {
        if (jsonDenseCaptionsResult.HasValue("values"))
        {
            JsonValue jsonDenseCaptionsResultValues = jsonDenseCaptionsResult["values"];

            for (int i = 0; i < jsonDenseCaptionsResultValues.Count(); ++i)
            {
                push_back(ContentCaption(jsonDenseCaptionsResultValues[i]));
            }
        }
    }

    /*! \endcond */
};

/// <summary>
/// Represents a single word that was detected in an image.
/// </summary>
/// <remarks>
/// Words consist of a contiguous sequence of characters.
/// For non-space delimited languages such as Chinese, Japanese, and Korean, each character is represented as its
/// own word.
/// </remarks>
class DetectedTextWord
{
public:
    /// <summary>
    /// The text detected within the bounds of this word.
    /// </summary>
    std::string Content;

    /// <summary>
    /// A score that represents the likelihood that this detection was accurate.
    /// </summary>
    /// <remarks>
    /// Confidence scores span the range of 0.0 to 1.0 (inclusive), with higher values indicating higher
    /// probability.
    /// </remarks>
    double Confidence = 0;

    /// <summary>
    /// A bounding polygon with points that enclose the word.
    /// </summary>
    /// <remarks>
    /// These points are polygon vertices, presented in clockwise order from the left (-180 degrees, inclusive)
    /// relative to the region's orientation.
    ///
    /// Coordinates are are in pixels, with (0,0) being the top-left of the source image.
    /// </remarks>
    std::vector<int32_t> BoundingPolygon;

private:
    /// <summary>
    /// Recorded to properly bound the substring lookup into the parent result's text.
    /// </summary>
    int32_t Offset = 0;

    /// <summary>
    /// Recorded to properly bound the substring lookup into the parent result's text.
    /// </summary>
    int32_t Length = 0;

public:
    /*! \cond PRIVATE */

    /// <summary>
    /// Construct the DetectedTextWord object from the appropriate JSON element in the service response.
    /// </summary>
    DetectedTextWord(JsonValue jsonWordResult)
    {
        Content = jsonWordResult["content"].AsString();
        Confidence = jsonWordResult["confidence"].AsDouble();
        Offset = (int32_t)jsonWordResult["span"]["offset"].AsInt();
        Length = (int32_t)jsonWordResult["span"]["length"].AsInt();

        for (int j = 0; j < jsonWordResult["boundingBox"].Count(); ++j)
        {
            int32_t pixel = (int32_t)jsonWordResult["boundingBox"][j].AsInt();
            BoundingPolygon.push_back(pixel);
        }
    }

    /*! \endcond */
};

/// <summary>
/// Represents a single, contiguous line of text as detected within an image.
/// </summary>
class DetectedTextLine
{
public:
    /// <summary>
    /// The text detected in this line.
    /// </summary>
    std::string Content;

    /// <summary>
    /// A bounding polygon with points that enclose this line of text.
    /// </summary>
    /// <remarks>
    /// These points are polygon vertices, presented in clockwise order from the left (-180 degrees, inclusive)
    /// relative to the region's orientation.
    ///
    /// Coordinates are are in pixels, with (0,0) being the top-left of the source image.
    /// </remarks>
    std::vector<int32_t> BoundingPolygon;

    /// <summary>
    /// A list of detected words associated with this line.
    /// </summary>
    std::vector<DetectedTextWord> Words;

    /*! \cond PRIVATE */

    /// <summary>
    /// Construct the DetectedTextLine object from the appropriate JSON element in the service response.
    /// </summary>
    DetectedTextLine(JsonValue jsonLineResult, JsonValue jsonWordsResult)
    {
        Content = jsonLineResult["content"].AsString();

        // At the moment we only have a single element in the "spans" array
        int32_t lineOffset = (int32_t)jsonLineResult["spans"][0]["offset"].AsInt();
        int32_t lineLength = (int32_t)jsonLineResult["spans"][0]["length"].AsInt();

        for (int i = 0; i < jsonLineResult["boundingBox"].Count(); ++i)
        {
            int32_t pixel = (int32_t)jsonLineResult["boundingBox"][i].AsInt();
            BoundingPolygon.push_back(pixel);
        }

        // Find all the words that belong to this line based on the offset and length
        for (int i = 0; i < jsonWordsResult.Count(); ++i)
        {
            int32_t wordOffset = (int32_t)jsonWordsResult[i]["span"]["offset"].AsInt();
            int32_t wordLength = (int32_t)jsonWordsResult[i]["span"]["length"].AsInt();

            if (wordOffset >= lineOffset && wordOffset + wordLength <= lineOffset + lineLength)
            {
                Words.push_back(DetectedTextWord(jsonWordsResult[i]));
            }
        }
    }

    /*! \endcond */
};

/// <summary>
/// Represents the text lines detected in an image.
/// </summary>
class DetectedText
{
public:
    /// <summary>
    /// The full list of all lines of text detected in an image.
    /// </summary>
    std::vector<DetectedTextLine> Lines;

    /*! \cond PRIVATE */

    /// <summary>
    /// Construct the DetectedTextBlocks object from the appropriate JSON element in the service response.
    /// </summary>
    DetectedText(JsonValue jsonReadResultPage0)
    {
        for (int i = 0; i < jsonReadResultPage0["lines"].Count(); ++i)
        {
            Lines.push_back(DetectedTextLine(jsonReadResultPage0["lines"][i], jsonReadResultPage0["words"]));
        }
    }

    /*! \endcond */
};

/// <summary>
/// Holds a single segmentation result image of PNG format and associated metadata.
/// </summary>
class SegmentationResult
{
public:
    /// <summary>
    /// Get the result image buffer in PNG format.
    /// </summary>
    /// <returns>
    /// A shared pointer to the image buffer.
    /// </returns>
    std::shared_ptr<uint8_t> GetImageBuffer()
    {
        return imageBuffer;
    };

    /// <summary>
    /// Get the size in bytes of the result image buffer.
    /// </summary>
    /// <returns>
    /// Buffer size in bytes.
    /// </returns>
    uint32_t GetImageBufferSize()
    {
        // Allocates new memory, copies the image buffer into it and returns a vector that owns the memory.
        return imageBufferSize;
    };

    /// <summary>
    /// Gets the image height in pixels.
    /// </summary>
    /// <returns>
    /// The image height in pixels.
    /// </returns>
    int32_t GetImageHeight()
    {
        return imageHeight;
    };
    
    /// <summary>
    /// Gets the image width in pixels.
    /// </summary>
    /// <returns>
    /// The image width in pixels.
    /// </returns>
    int32_t GetImageWidth()
    {
        return imageWidth;
    }

public:
    /*! \cond PRIVATE */

    /// <summary>
    /// Construct the SegmentationResult object from response buffer
    /// </summary>
    SegmentationResult(
        std::shared_ptr<uint8_t> bufferPtr,
        uint32_t bufferSize,
        int32_t height,
        int32_t width)
    {
        imageBuffer = bufferPtr;
        imageBufferSize = bufferSize;
        imageHeight = height;
        imageWidth = width;
    }

    /*! \endcond */

private:
    std::shared_ptr<uint8_t> imageBuffer;
    uint32_t imageBufferSize = 0;
    int32_t imageHeight = 0;
    int32_t imageWidth = 0;
};

/// <summary>
/// Represents the outcome of an Image Analysis operation.
/// </summary>
/// <remarks>
/// Always start by calling GetReason to determine if the analysis
/// operation was successful or not.
/// When an analysis operation is successful, applicable Get methods in this object
/// will return results based on the selected features (ImageAnalysisOptions::SetFeatures)
/// or custom-trained model (ImageAnalysisOptions::SetModelName).
/// These results are parsed from the service JSON response.
/// Other Get methods will return a Nullable result objects with no value.
/// We use the Nullable class to hold a result, as a convenient container that can
/// indicate if a result has value (see Nullable::HasValue).
/// Call ImageAnalysisResultDetails::FromResult to get access to additional
/// information about the result, such as the raw JSON response.
/// When analysis operation failed, call ImageAnalysisErrorDetails::FromResult
/// to get access to additional information about the error.
/// </remarks>
class ImageAnalysisResult : public Core::Details::ResultBase<ImageAnalysisCoreResultReason, int, ImageAnalysisResultProperties>
{
private:

    using BaseResult = ResultBase<ImageAnalysisCoreResultReason, int, ImageAnalysisResultProperties>;

public:

    /// <summary>
    /// Destructs an instance of the ImageAnalysisResult class.
    /// </summary>
    ~ImageAnalysisResult() = default;

    /// <summary>
    /// Gets a value indicating why this result was generated.
    /// </summary>
    /// <returns>
    /// Reason of this result.
    /// </returns>
    ImageAnalysisResultReason GetReason() const
    { 
        return reason;
    }

    /// <summary>
    /// Gets a collection of additional properties on the result.
    /// </summary>
    const ImageAnalysisResultProperties& Properties;

    /// <summary>
    /// Gets the model version used by the Image Analysis Service to create this result.
    /// </summary>
    /// <returns>
    /// The version of the model.
    /// </returns>
    Nullable<std::string> GetModelVersion()
    { 
        return modelVersion;
    }

    /// <summary>
    /// Gets the height, in pixels, of the analyzed image.
    /// </summary>
    /// <returns>
    /// A Nullable object holding an int32_t value representing the height, or an empty value
    /// </returns>
    /// <remarks>
    /// Only populated if the image was analyzed successfully
    /// </remarks>
    Nullable<int32_t> GetImageHeight()
    {
        return imageHeight;
    }

    /// <summary>
    /// Gets the width, in pixels, of the analyzed image. Only populated
    /// if the image was analyzed successfully.
    /// </summary>
    /// <returns>
    /// A Nullable object holding an int32_t value representing the width, or an empty value
    /// </returns>
    Nullable<int32_t> GetImageWidth()
    {
        return imageWidth;
    }

    /// <summary>
    /// Gets a generated phrase that describes the content of the analyzed image.
    /// </summary>
    /// <remarks>
    /// This result will only be populated if ImageAnalysisFeature::Caption was included
    /// while calling ImageAnalysisOptions::SetFeatures.
    /// </remarks>
    /// <returns>
    /// A Nullable object holding the ContentCaption result or an empty value
    /// </returns>
    Nullable<ContentCaption> GetCaption()
    {
        return contentCaption;
    }

    /// <summary>
    /// Gets up to 10 generated phrases, the first describing the content of the whole image,
    /// and the others describing the content of different regions of the image.
    /// </summary>
    /// <remarks>
    /// This result will only be populated if ImageAnalysisFeature::DenseCaptions was included
    /// while calling ImageAnalysisOptions::SetFeatures.
    /// </remarks>
    /// <returns>
    /// A Nullable object holding the DenseCaptions result or an empty value
    /// </returns>
    Nullable<DenseCaptions> GetDenseCaptions()
    {
        return denseCaptions;
    }

    /// <summary>
    /// Gets a list of detected objects from the analyzed image.
    /// </summary>
    /// <remarks>
    /// This result will only be populated if ImageAnalysisFeature::Objects was included
    /// while calling ImageAnalysisOptions::SetFeatures.
    /// </remarks>
    /// <returns>
    /// A Nullable object holding the DetectedObjects result or an empty value
    /// <returns>
    Nullable<DetectedObjects> GetObjects()
    {
        return detectedObjects;
    }

    /// <summary>
    /// Gets a list of detected objects from the analyzed image using
    /// the provided custom-trained model.
    /// </summary>
    /// <remarks>
    /// This result may be populated if ImageAnalysisOptions::SetModelName
    /// was called, and the provided custom model was trained to detect objects.
    /// </remarks>
    /// <returns>
    /// A Nullable object holding the DetectedObjects result or an empty value
    /// <returns>
    Nullable<DetectedObjects> GetCustomObjects()
    {
        return customDetectedObjects;
    }

    /// <summary>
    /// Gets a collection of extracted textual lines and words from an analyzed image.
    /// </summary>
    /// <remarks>
    /// This result will only be populated if ImageAnalysisFeature::Text was included
    /// while calling ImageAnalysisOptions::SetFeatures.
    /// </remarks>
    /// <returns>
    /// A Nullable object holding the DetectedText result or an empty value.
    /// </returns>
    Nullable<DetectedText> GetText()
    {
        return detectedText;
    }

    /// <summary>
    /// Gets a list of content tag detections from the analyzed image.
    /// </summary>
    /// <remarks>
    /// This result will only be populated if ImageAnalysisFeature::Tags was included
    /// while calling ImageAnalysisOptions::SetFeatures.
    /// </remarks>
    /// <returns>
    /// A Nullable object holding the ContentTags result or an empty value.
    /// </returns>
    Nullable<ContentTags> GetTags()
    {
        return contentTags;
    }

    /// <summary>
    /// Gets a list of content tag detections from the analyzed image using
    /// the provided custom-trained model.
    /// </summary>
    /// <remarks>
    /// This result may be populated if ImageAnalysisOptions::SetModelName
    /// was called, and the provided custom model was trained to detect tags.
    /// </remarks>
    /// <returns>
    /// A Nullable object holding the ContentTags result or an empty value.
    /// </returns>
    Nullable<ContentTags> GetCustomTags()
    {
        return customContentTags;
    }

    /// <summary>
    /// Gets a list of detected people from an analyzed image.
    /// </summary>
    /// <remarks>
    /// This result will only be populated if ImageAnalysisFeature::People was included
    /// while calling ImageAnalysisOptions::SetFeatures.
    /// </remarks>
    /// <returns>
    /// A Nullable object holding the DetectedPeople result or an empty value.
    /// </returns>
    Nullable<DetectedPeople> GetPeople()
    {
        return detectedPeople;
    }

    /// <summary>
    /// Gets a list of suggested image crop operations of the analyzed
    /// image at the desired aspect ratios (if provided)
    /// </summary>
    /// <remarks>
    /// Also knows as SmartCrops. These cropping suggestions preserve as much content 
    /// as possible while achieving the specified aspect ratios (if provided).
    ///
    /// This result will only be populated if ImageAnalysisFeature::CropSuggestions was included
    /// while calling ImageAnalysisOptions::SetFeatures.
    ///
    /// Optionally, specify one or more desired cropping aspect ratios by calling 
    /// ImageAnalysisOptions::SetCroppingAspectRatios. If CroppingAspectRatios is not set,
    /// the Service will return one crop suggestion with an aspect ratio it sees fit.
    /// </remarks>
    /// <returns>
    /// A Nullable object holding the CropSuggestions result or an empty value.
    /// </returns>
    Nullable<CropSuggestions> GetCropSuggestions()
    {
        return cropSuggestions;
    }

    /// <summary>
    /// Get the result of the image segmentation operation.
    /// </summary>
    /// <remarks>
    /// This result will only be populated if ImageAnalysisOptions::SetSegmentationMode()
    /// was called with either ImageSegmentationMode::BackgroundRemoval or
    /// ImageSegmentationMode::ForegroundMatting.
    /// </remarks>
    /// <returns>
    /// A Nullable object holding the SegmentationResult or an empty value.
    /// </returns>
    Nullable<SegmentationResult> GetSegmentationResult()
    {
        return segmentationResult;
    }

protected:
     /*! \cond PROTECTED */
    static std::shared_ptr<ImageAnalysisResult> FromHandle(AZAC_HANDLE handle)
    {
        auto ptr = new ImageAnalysisResult(handle);
        return std::shared_ptr<ImageAnalysisResult>(ptr);
    }

    explicit ImageAnalysisResult(AZAC_HANDLE result) :
        ResultBase(result),
        Properties(GetProperties())
    {
        reason = GetCombinedReason();
        ParseResponse();
    }

    explicit operator AZAC_HANDLE() { return AI::Core::Details::ProtectedAccess<BaseResult>::HandleFromPtr(this); }
    /*! \endcond */

private:

    AZAC_DISABLE_DEFAULT_CTORS(ImageAnalysisResult);

    // True of the analyzed response result was already parsed
    bool responseParsed = false;

    // Values parsed from the JSON response, if jsonParsed = true.
    // These values will be present in every HTTP 200 OK response.
    // The Nullable object will be empty if the service returned an error.
    Nullable<std::string> modelVersion;
    Nullable<int32_t> imageHeight;
    Nullable<int32_t> imageWidth;

    // Values parsed from the JSON response, if responseParsed = true.
    // These are optional, depending on the visual features settings. Hence we use Nullable<>
    // values. The Nullable object will be empty if the value did not exist in the JSON response,
    // or if the service returned an error.
    Nullable<ContentTags> contentTags;
    Nullable<ContentTags> customContentTags;
    Nullable<ContentCaption> contentCaption;
    Nullable<DenseCaptions> denseCaptions;
    Nullable<DetectedObjects> detectedObjects;
    Nullable<DetectedObjects> customDetectedObjects;
    Nullable<DetectedText> detectedText;
    Nullable<DetectedPeople> detectedPeople;
    Nullable<CropSuggestions> cropSuggestions;

    // Values parsed from the binary response, if responseParsed = true.
    Nullable<SegmentationResult> segmentationResult;

    // The reason returned on this public result object, which may be different than ImageAnalysisCoreResultReason
    ImageAnalysisResultReason reason;

    // Parse the HTTP response payload of a successful analyzed call (either JSON or binary buffer), and
    // any other associated result properties. All the relevant result objects will be
    // created and populated. This only needs to be called once.
    void ParseResponse()
    {
        if (responseParsed)
        {
            return;
        }

        if (reason == ImageAnalysisResultReason::Analyzed)
        {
            std::string jsonResult= Properties.Get("image.analysis.result.json");

            if (!jsonResult.empty())
            {
                ParseJsonResponse(jsonResult);
            }
            else
            {
                ParseBinaryResponse();
            }
        }

        responseParsed = true;
    }

    // Parse the binary response payload from the service, following a 'segment' call to the service
    void ParseBinaryResponse()
    {
        int32_t responseImageWidth = static_cast<int32_t>(std::stoi(Properties.Get("result.segmentation.image.width")));
        int32_t responseImageHeight = static_cast<int32_t>(std::stoi(Properties.Get("result.segmentation.image.height")));
        uint32_t responseImageSize = 0;
        // Note that a call to Properties.GetBinary() allocates new memory, copies the binary data and returns a shared pointer to it.
        std::shared_ptr<uint8_t> responseImageBuffer = Properties.GetBinary("result.segmentation.image.buffer", &responseImageSize);

        if (responseImageBuffer != nullptr && responseImageSize > 0)
        {
            segmentationResult = SegmentationResult(responseImageBuffer, responseImageSize, responseImageHeight, responseImageWidth);
        }
    }

    // Parse the JSON response payload from the service, following an 'analyze' call to the service
    void ParseJsonResponse(std::string& jsonResult)
    {
        JsonValue jsonValueResults = JsonValue::Parse(jsonResult.c_str(), jsonResult.size());

        if (jsonValueResults.HasValue("modelVersion"))
        {
            modelVersion = jsonValueResults["modelVersion"].AsString();
        }

        if (jsonValueResults.HasValue("metadata"))
        {
            imageHeight = (int32_t)(jsonValueResults["metadata"]["height"].AsInt());
            imageWidth = (int32_t)(jsonValueResults["metadata"]["width"].AsInt());
        }

        if (jsonValueResults.HasValue("tagsResult"))
        {
            contentTags = ContentTags(jsonValueResults["tagsResult"]);
        }

        if (jsonValueResults.HasValue("objectsResult"))
        {
            detectedObjects = DetectedObjects(jsonValueResults["objectsResult"]);
        }

        if (jsonValueResults.HasValue("captionResult"))
        {
            contentCaption = ContentCaption(jsonValueResults["captionResult"], imageWidth.ValueOr(0), imageHeight.ValueOr(0));
        }

        if (jsonValueResults.HasValue("denseCaptionsResult"))
        {
            denseCaptions = DenseCaptions(jsonValueResults["denseCaptionsResult"]);
        }

        if (jsonValueResults.HasValue("readResult"))
        {
            detectedText = DetectedText(jsonValueResults["readResult"]["pages"][0]);
        }

        if (jsonValueResults.HasValue("peopleResult"))
        {
            detectedPeople = DetectedPeople(jsonValueResults["peopleResult"]);
        }

        if (jsonValueResults.HasValue("smartCropsResult"))
        {
            cropSuggestions = CropSuggestions(jsonValueResults["smartCropsResult"]);
        }

        if (jsonValueResults.HasValue("customModelResult"))
        {
            JsonValue customModelResult = jsonValueResults["customModelResult"];

            if (customModelResult.HasValue("tagsResult"))
            {
                customContentTags = ContentTags(customModelResult["tagsResult"]);
            }

            if (customModelResult.HasValue("objectsResult"))
            {
                customDetectedObjects = DetectedObjects(customModelResult["objectsResult"]);
            }
        }
    }

    // Private helper method. Converts the result reason and (if applicable) stop reason on a result into the
    // combined, public representation of a result reason. We don't separate stop and result reasons in the public
    // surface because IA has no concept of 'stop' yet.
    ImageAnalysisResultReason GetCombinedReason()
    {
        const ImageAnalysisResultProperties& properties = BaseResult::GetProperties();

        // Error check #1: if an "error.reason" exists -- well, it's an error
        if (!properties.Get("error.reason", "").empty())
        {
            return ImageAnalysisResultReason::Error;
        }

        // Error check #2 -- if "session.stopped.reason" exists and it converts to "error," it's an error
        if (properties.Get("session.stopped.reason", "") == std::to_string(static_cast<int32_t>(Azure::AI::Vision::ImageAnalysis::Details::ImageAnalysisCoreStopReason::Error)))
        {
            return ImageAnalysisResultReason::Error;
        }

        // If it's not an error, the "result.reason" should indicate it's a successful analysis result
        if (properties.Get("result.reason", "") == std::to_string(static_cast<int32_t>(Azure::AI::Vision::ImageAnalysis::Details::ImageAnalysisCoreResultReason::Analyzed)))
        {
            return ImageAnalysisResultReason::Analyzed;
        }

        // If we got this far and none of the above fit? Something went wrong
        return ImageAnalysisResultReason::Error;
    }

    // These base class methods are not needed at the moment. Make sure they cannot be used.
    template<class T = std::string> AI::Core::Details::enable_if_w_or_string_t<T> GetSessionId() const = delete;
    template<class T = std::string> AI::Core::Details::enable_if_w_or_string_t<T> GetResultId() const = delete;
    uint64_t GetAssociatedMediaPosition() const = delete;
    std::shared_ptr<Input::Frames::FrameReader> GetAssociatedFrameReader() = delete;
    std::shared_ptr<Input::Frames::FrameSetReader> GetAssociatedFrameSetReader() = delete;
};

} } } } // Azure::AI::Vision::ImageAnalysis
